import os, pkg_resources, zipfile, shutil
from io import BytesIO
from lxml import etree


def create_schema_documentation(xml_schema_string , json_schema_string, output_folder, output_name, glossary_string=None):
    """
    Generates documentation for the provided LIXI schema. Documentation is based on a custom version of readthedocs [https://docs.readthedocs.io/en/rel/theme.html] theme.

    Args:
        xml_schema_string (:obj:`str`, required): LIXI XML schema provided as string.
        json_schema_string (:obj:`str`, required): LIXI XML schema provided as string.
        output_folder (:obj:`str`, required): Path to write the produced lixi schema documentation to.
        outputname (:obj:`str`, required): Custom schema name used in namespace.
        glossary_string (:obj:`str`, optional): LIXI Glossary update provided as a string. Defaults to None.
            
    Returns:
        Outputs the custom documentation for a schema to the specified folder. 
    """   
    
    if output_folder!=None:
        

        def copytree(src, dst, symlinks=False, ignore=None):
            
            if os.path.exists(dst):
                shutil.rmtree(dst)
                os.mkdir(dst)
            else:    
                os.mkdir(dst)
            
            for item in os.listdir(src):
                s = os.path.join(src, item)
                d = os.path.join(dst, item)
                if os.path.isdir(s):
                    shutil.copytree(s, d, symlinks, ignore)
                else:
                    if item not in ['README.md', 'create_documentation.py', '.gitlab-ci.yml', '.gitignore']:
                        shutil.copy2(s, d)    
        
        # First create the directory
        output_path = os.path.join(output_folder, output_name) 
        
        # Copy all documentation files from lib to provided path
        docs_location = pkg_resources.resource_filename(__name__, "documentation")
        copytree(docs_location, output_path)
        
        # Write data to file to display correct info
        data = "var schemaString = `" + xml_schema_string + "`"
        writepath = os.path.join(output_path, 'js', 'LIXI-Master-Schema.js')
        with open(writepath, 'w+') as file:
            data = file.write(data)
            
        if glossary_string !=None:
            data = "var glossaryString = `" + glossary_string + "`"
            writepath = os.path.join(output_path, 'js', 'LIXI_Glossary.js')
            with open(writepath, 'w+') as file:
                data = file.write(data)
            
        return True        
        
    else:
        inMemoryOutputFile = BytesIO()
        zipFile = zipfile.ZipFile(inMemoryOutputFile, 'w', zipfile.ZIP_DEFLATED) 
        path = pkg_resources.resource_filename(__name__, "documentation")
    
        for root, dirs, files in os.walk(path):
            for file in files:
                if file not in ['README.md', 'create_documentation.py', '.gitlab-ci.yml', '.gitignore']: 
                    path_after_doc = output_name + root.split('documentation')[1]
                    zipFile.write(os.path.join(root, file), os.path.join(path_after_doc, file))
    
        text = "var schemaString = vkbeautify.xml(`" + xml_schema_string + "`)"
        zipFile.writestr(os.path.join(output_name,'js','LIXI-Master-Schema.js'), text)
        
        if glossary_string !=None:
            text = "var glossaryString = vkbeautify.xml(`" + glossary_string + "`)"
            zipFile.writestr(os.path.join(output_name,'js','LIXI_Glossary.js'), text)        
    
        if json_schema_string:
            zipFile.writestr(os.path.join(output_name + '.json'), json_schema_string)
        else:
            zipFile.writestr(os.path.join(output_name + '.xsd'), xml_schema_string)
        
        zipFile.close()
    
        return inMemoryOutputFile.getvalue()


        
        


import os, re, filecmp, sys, datetime, html, inspect, lxml, pkg_resources, glob, zipfile, io, xml.dom.minidom as dome
from operator import attrgetter
from shutil import copyfile
from lxml import etree
from lxml import builder

ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}


def apply_xslt(xmlFile, xsltFile):
    """ Applys a xslt rule file to a XML message.
    
    Args:
        xmlFile (:obj:`lxml.etree`, required): A XML Message.
        xsltFile (:obj:`lxml.etree`, required): A XSLT transform rules file.
    
    Returns:
        A string of result XML.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    xslt = etree.parse(xsltFile)
    transform = etree.XSLT(xslt)
    newdom = transform(xmlFile)
    outstr = etree.tostring(
        newdom, pretty_print=True, xml_declaration=True, encoding="UTF-8"
    )

    return outstr


def transform_xslt(sample_etree, xslt_folder_path, transaction_type, from_version_str, to_version_str):
    """ Transforms a LIXI Message instance to an older/newer version, based on the schema version specified, of the LIXI Message instance.
    
    Args:
        sample_etree (:obj:`str`, required): A LIXI Message provided as a string. Defaults to None.
        xslt_folder_path (:obj:`bool`, required): A path to a folder containing LIXI transformation (.xslt) files. 
        transaction_type (:obj:`bool`, required): The transaction type of a LIXI schema. Should be one of 'ACC', 'CAL', 'CDA', 'CNZ', 'DAS', 'LMI', 'SVC', 'VAL'. Defaults to None.
        from_version_str (:obj:`str`, required): The version of the LIXI schema to be transformed to. Should be in the format of '2.6.24'. 
        to_version_str (:obj:`str`, required): The version of the LIXI schema to be transformed to. Should be in the format of '2.6.24'. 

    Returns:
        Transformed LIXI Message instance ONLY if return_instance is set to True.

    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    def sort_versions(tranforms):
        temp_list = []

        def num_keyfn(x):
            parts = x.split("_")
            return tuple(int(x) if x.isnumeric else x for x in parts)

        for num in sorted(tranforms, key=num_keyfn):
            temp_list.append(num)

        return temp_list

    # Determine if its a backwards or forwards transformation (assumption, version provided by user is of the form '#.#.#')
    to_version = to_version_str.split(".")
    from_version = from_version_str.split(".")

    if int(to_version[0]) == int(from_version[0]):
        if int(to_version[1]) == int(from_version[1]):
            if int(to_version[2]) == int(from_version[2]):
                return sample_etree, ""
            elif int(to_version[2]) > int(from_version[2]):
                direction = "forwards"
            elif int(to_version[2]) < int(from_version[2]):
                direction = "backwards"
        elif int(to_version[1]) > int(from_version[1]):
            direction = "forwards"
        elif int(to_version[1]) < int(from_version[1]):
            direction = "backwards"
    elif int(to_version[0]) > int(from_version[0]):
        direction = "forwards"
    elif int(to_version[0]) < int(from_version[0]):
        direction = "backwards"

    # Vars help order the final list correctly
    if direction == "forwards":
        greater_version_no = (
            "LIXI-" + transaction_type + "-" + to_version_str.replace(".", "_")
        )
        lesser_version_no = (
            "LIXI-" + transaction_type + "-" + from_version_str.replace(".", "_")
        )
    else:
        greater_version_no = (
            "LIXI-" + transaction_type + "-" + from_version_str.replace(".", "_")
        )
        lesser_version_no = (
            "LIXI-" + transaction_type + "-" + to_version_str.replace(".", "_")
        )

    # The complete version lists
    tranforms = []
    #for filename in glob.glob(
        #os.path.join(pkg_resources.resource_filename(__name__, "xslts"), "*.xsl")
    #):
    for filename in glob.glob(
        os.path.join(xslt_folder_path, "*.xsl")
    ):        
        if "warnings" not in filename and transaction_type in filename:
            filename = re.search(
                "LIXI-([a-zA-Z0-9_-]*).xsl", filename.replace("LIXI", "", 1)
            ).group(1)
            filename = filename.replace(transaction_type + "-", "")
            if filename not in tranforms:
                tranforms.append(filename)
    tranforms = sort_versions(tranforms)
    tranforms_versions = [
        ("LIXI-" + transaction_type + "-" + x.replace(".", "_")) for x in tranforms
    ]

    # Create a list of only required version/transaction strings and in the correct order.
    count_forwards = []
    count_backwards = []
    tranforms_list = []
    check = False

    for version in tranforms_versions:
        if version == lesser_version_no or check == True:
            count_forwards.append(version)
            check = True

    check = False
    count_forwards = list(reversed(count_forwards))
    for version in count_forwards:
        if version == greater_version_no or check == True:
            count_backwards.append(version)
            check = True

    if direction == "forwards":
        tranforms_list = list(reversed(count_backwards))
    else:
        tranforms_list = count_backwards

    tranforms_forwards, tranforms_backwards = (
        count_backwards,
        list(reversed(count_backwards)),
    )

    if direction == "forwards":
        tranforms_forwards, tranforms_backwards = (
            tranforms_backwards,
            tranforms_forwards,
        )

    # The actual transformation of the files from one version to all prev versions
    step = 0

    temp_prev_step_xml = sample_etree
    temp_current_step_xml = None
    temp_step_warnings_xml = etree.Element(
        "{http://www.w3.org/2001/XMLSchema-instance}warnings"
    )

    for first, second in zip(tranforms_list, tranforms_list[1:]):

        #archive = zipfile.ZipFile(
            #os.path.join(
                #pkg_resources.resource_filename(__name__, "xslts"),
                #first + "-" + second + ".zip",
            #),
            #"r",
        #)
        #stylesheet = io.BytesIO(archive.read(first + "-" + second + ".xsl"))
        stylesheet = os.path.join(xslt_folder_path,first + "-" + second + ".xsl")
        
        #archive = zipfile.ZipFile(
            #os.path.join(
                #pkg_resources.resource_filename(__name__, "xslts"),
                #first + "-" + second + "-warnings.zip",
            #),
            #"r",
        #)
        #warningsheet = io.BytesIO(archive.read(first + "-" + second + "-warnings.xsl"))
        warningsheet = os.path.join(xslt_folder_path, first + "-" + second + "-warnings.xsl")
        
        # schemafile   = os.path.join(xsd_folder_path, second + "-Annotated.xsd")

        # print(stylesheet + "\t\t" + "-step-" + str(step + 1), end="")

        # do the XSL transformation
        temp_current_step_xml = etree.fromstring(
            apply_xslt(temp_prev_step_xml, stylesheet)
        )

        # step
        comment = etree.Comment(" Step " + str(step) + " ")
        temp_step_warnings_xml.append(comment)

        # create warnings for this transformation
        temp_current_step_warnings_xml = etree.fromstring(
            apply_xslt(temp_prev_step_xml, warningsheet)
        )
        for child in temp_current_step_warnings_xml:
            temp_step_warnings_xml.append(child)

        # Validate the resulting XML against the correct schema
        # schema_etree = etree.parse(schemafile)
        # schema = etree.XMLSchema(schema_etree)
        # valid = schema.validate(temp_current_step_xml)
        # valid = True
        # if valid:
        # print("\t...valid")

        # else:
        # print("\n")
        # for error in schema.error_log:
        # print("ERROR ON LINE "+str(error.line)+": "+str(error.message.encode("utf-8"))+".")
        # raise Exception("ERROR: tranformed message is not valid, "+str(step + 1)+".")

        temp_prev_step_xml = temp_current_step_xml
        step = step + 1

    xml_string = etree.tostring(temp_step_warnings_xml).decode()
    xml_string = xml_string.strip().replace(">\n", ">").replace("\n<", "<")
    reparsed = dome.parseString(xml_string)
    all_warnings = reparsed.toprettyxml(indent="  ")

    return temp_current_step_xml, all_warnings



from lxml import isoschematron
from lxml import etree


def validate(xml_etree, schematron_etree):
    """ Validates a LIXI message against a Schematron business rules file.
    
    Args:
        xml_etree (:obj:`lxml.etree`, required): message to check provided as a Python Etree.
        schematron_etree (:obj:`lxml.etree`, required): Schematron rules provided as a Python Etree.
        error_log (:obj:`list`): Validation error object as a Python List.
    
    Result:
        result (:obj:`bool`): Indicates if validation was successful.
        message (:obj:`str`): Validation message.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """
    
    class Error_Log:
        """Represents a Error when validating with a schematron file.
            
        """
        def __init__(self, message, test, message_line, message_location):
            """Creates a LIXI message XML.
        
            Args:
                message (:obj:`str`): Error string of the message.
                test (:obj:`str`) : Test that caused the Error.
                message_line (:obj:`str`): Line number in the input message.
                message_location (:obj:`str`): Location xpath in the input message.

            """
    
            self.message = str.strip(message)
            self.test = test
            self.message_line = str(message_line)
            self.message_location = message_location             

    schematron = isoschematron.Schematron(schematron_etree, store_report=True)
    validation_result = schematron.validate(xml_etree)

    report = schematron.validation_report

    if validation_result:
        return True, "Message schematron validation is successful.", None
    else:
        err_message = "Schematron validation failed."
        err_message += "\n\nInvalid Assertions:"
        
        list_of_error = []
        
        for fail in report.xpath(
            "//svrl:failed-assert",
            namespaces={"svrl": "http://purl.oclc.org/dsdl/svrl"},
        ):
            
            result =  xml_etree.xpath(str(fail.get("location")))
            if len(result)>0:
                line = str(result[0].sourceline)            
            else:
                line = "X"
            
            err_message = (
                err_message
                + "\n    Error On Line "
                + line
                + ": "
                + str(fail.getchildren()[0].text).strip()
            )
            
            list_of_error.append(Error_Log(fail.getchildren()[0].text, fail.get('test'), line, str(fail.get("location")) ))

        return False, err_message, list_of_error



import os, sys, re, io, copy
from lxml import etree

'''
e.g.

Create_Uncomplextypes_Schema.py "../../Schemas/LIXI-CAL-2_4_2-Annotated.xsd" "./Untyped_Schemas/LIXI-CAL-2_4_2-Annotated.xsd"

'''

ns = {'xs': 'http://www.w3.org/2001/XMLSchema',
      'lx': 'lixi.org.au/schema/appinfo_elements',
      'li': 'lixi.org.au/schema/appinfo_instructions'}
    
def load(path):
    '''
    Read Schema from file
    '''
    #with open(path) as file:
        #data = file.read().
        #return etree.fromstring(data)
        
    
    f = io.open(path, mode="r", encoding="utf-8")
    data =  f.read()
    
    data = bytes(bytearray(data, encoding='utf-8'))
    
    parser = etree.XMLParser(remove_blank_text=True)

    
    return etree.fromstring(data, parser)
    
    
def save(filename, schema):
    '''
    Write Schema to file
    '''
    schema_string = etree.tostring(schema, pretty_print=True, encoding='UTF-8')
   
    
    with open(filename, 'wb') as file:
        file.write(schema_string)
    
    parser    = etree.XMLParser(remove_blank_text=True)
    root      = etree.parse(filename, parser).getroot()
    stringout = etree.tostring(root, pretty_print=True, encoding='UTF-8')
    
    with open(filename, 'wb') as file:
        file.write(stringout)
        
        
def clone(item):
    return copy.deepcopy(item)

def get_all_lixi_elements (schema):
    xpath = "//xs:element"
    return schema.xpath(xpath, namespaces=ns)

def get_all_lixi_items (schema):
    xpath = ".//xs:element|.//xs:attribute"
    return schema.xpath(xpath, namespaces=ns)

def get_all_lixi_paths (schema):
    xpath = "//lx:path"
    return schema.xpath(xpath, namespaces=ns)

def get_all_lixi_complex_types (schema):
    xpath = "/xs:schema/xs:complexType"
    return schema.xpath(xpath, namespaces=ns)

def get_all_lixi_elements_using_a_complex_type (schema, complextype):
    name = names([complextype])[0]
    xpath = "//xs:element[@type='"+ name +"']"
    return schema.xpath(xpath, namespaces=ns)

def get_package_element (schema):
    xpath = "/xs:schema/xs:element[@name='Package']"
    return schema.xpath(xpath, namespaces=ns)[0]

def get_item (schema, path):
    ''' given a path, get the schema item '''
    xpath = "//lx:path[text()='"+path+"']/../../.."
    return schema.xpath(xpath, namespaces=ns)[0]

def get_decendants (node):
    ''' given an lxml schema node, return a list of lixi items (elements, attributes, enumerations) that are decendants'''
    return node.xpath('.//xs:element|.//xs:attribute|.//xs:enumeration', namespaces=ns)
    
    



def intersection(one, two):
    '''
    given two list of elements, return the items found in both lists
    '''
    a = set(one)
    b = set(two)
    return a.intersection(b)    


def urls(items):
    '''
    given a list of schema item, return a list of the relative URL of the item in the documentation
    '''
    urls = []

    for item in items:
        tag = tags([item])[0]
        path = paths([item])[0]
        name = names([item])[0]

        url = '../' + tag + '/index.html#' + path


        if tag == 'enumeration':
            x = len('.' + name) 
            path = path[:-x]
            url = '../simpletype/index.html#' + path + '-simpletype'

        else:
            url = '../' + tag +  '/index.html#' + path    

        urls.append(url)

    return urls


def subschema(schema, transaction):
    '''
    Given a schema and a transaction type (e.g. 'DAS'), return all schema items in the subschema.
    '''
    return schema.xpath('//li:transactions[contains(.,"' + transaction + '")]/../../..', namespaces=ns)


def subschema_exclusive(schema, transaction):
    '''
    Given a schema and a transaction type (e.g. 'DAS'), return all schema items in only found in this subschema.
    '''
    return schema.xpath("//li:transactions[text()='"+transaction+"']/../../..", namespaces=ns)


def deprecated(schema):
    '''
    return the items that are being deprecated
    '''
    return schema.xpath('//xs:documentation[contains(.,"deprecated")]/../..', namespaces=ns)
    
 
def references(schema):
    '''
    return the items that are cross references
    '''
    return schema.xpath('//xs:attribute[contains(.,"x_")]', namespaces=ns)

def get_name (item):
    if item.tag == '{http://www.w3.org/2001/XMLSchema}enumeration':
        return item.attrib['value']
    else: 
        return item.attrib['name']

def names(items):
    '''
    given a list of lxml schema items, return a list of names (or values for enumerations)
    '''
    result = []
    for item in items:
        if item.tag == '{http://www.w3.org/2001/XMLSchema}enumeration':
            result.append(item.attrib['value'])
        else: 
            result.append(item.attrib['name'])
    return result

def get_path (item):
    
    return item.getchildren()[0].getchildren()[1].getchildren()[0].text

def paths(items):
    '''
    Given a list of lxml schema items, return a list of paths
    '''
    result = []
    for item in items:
        result.append(item.getchildren()[0].getchildren()[1].getchildren()[0].text)
    return result


def labels(items):
    '''
    Given a list of lxml schema items, return a list of labels
    '''
    result = []
    for item in items:
        result.append(item.getchildren()[0].getchildren()[1].getchildren()[1].text)
    return result


def transactions(items):
    '''
    Given a list of lxml schema items, return a list of transaction tags
    '''
    result = []
    for item in items:
        result.append(item.getchildren()[0].getchildren()[1].getchildren()[2].text)
    return result


def items (schema, paths):
    '''
    Given a list of paths, return a list of lxml objects 
    '''
    result=[]
    for path in paths:
        path = path.replace("'","&apos;")
        items = schema.xpath("//lx:path[text()='"+path+"']/../../..", namespaces=ns)
        if len(items) > 0:
            result.append(items[0])
    return result


def types(items):
    '''
    Given a list of lxml schema items, return a list of types
    '''
    result = []
    for item in items:
        if 'type' in item.attrib:
            result.append(item.attrib['type'])
        else:
            result.append("")
    return result


def uses(items):
    '''
    Given a list of lxml schema items, return a list of uses
    '''
    result = []
    for item in items:
        if 'use' in item.attrib:
            result.append(item.attrib['use'])
        else:
            result.append("")
    return result


def item_types(items):
    '''
    Given a list of lxml schema items, return a list of xml types ("element", "attribute", "enumeration, "simpleType", or "complextType")
    '''
    result = []
    for item in items:
        tag = item.tag.replace("{http://www.w3.org/2001/XMLSchema}",'')
        result.append(tag)
    return result


def documentations(items):
    '''
    given a list of lxml schema items, return a list of paths
    '''
    result = []
    for item in items:
        result.append(item.getchildren()[0].getchildren()[0].text)
    return result


def minOccurs (items):
    '''
    given a list of lxml schema items, return a list of minOccurs
    '''
    result = []
    for item in items:
        if 'minOccurs' in item.attrib:
            result.append(item.attrib['minOccurs'])
        else:
            result.append("1")
    return result

def maxOccurs (items):
    '''
    given a list of lxml schema items, return a list of maxOccurs
    '''
    result = []
    for item in items:
        if 'maxOccurs' in item.attrib:
            result.append(item.attrib['maxOccurs'])
        else:
            result.append("")
    return result


def gettype(schema, name):
    '''
    Given a schema and a type name, return the complex type element, or simple type elememnt 
    '''
    return schema.xpath('//xs:complexType[@name="' + name + '"]', namespaces=ns)[0]


def getchildren(item):
    '''
    Given an lxml lixi element, return the lixi child elements as lxml objects
    '''
    children = []
    
    if item.tag == '{http://www.w3.org/2001/XMLSchema}complexType':
        child_elements = item.getchildren()[1].getchildren()[1:]
        for child_element in child_elements:
            children.append(child_element)
        
    else:    
    
        item.getchildren()
        if len(item.getchildren()) > 1:
            child_elements = item.getchildren()[1][0].getchildren()
            for child_element in child_elements:
                if (child_element.tag == '{http://www.w3.org/2001/XMLSchema}element'):
                    children.append(child_element)
                
    return children


def getchildren_include_choices(item):
    '''
    Given an lxml lixi element, return the lixi child elements as lxml objects
    '''
    children = []

    child_elements = item.xpath('./xs:complexType/xs:choice/xs:sequence/xs:choice/xs:element|./xs:complexType/xs:choice/xs:sequence/xs:choice/xs:sequence/xs:element|./xs:complexType/xs:choice/xs:sequence/xs:element|./xs:complexType/xs:sequence/xs:sequence/xs:element|./xs:complexType/xs:sequence/xs:element|./xs:complexType/xs:sequence/xs:choice/xs:element|./xs:complexType/xs:choice/xs:element|./xs:complexType/xs:sequence/xs:choice/xs:sequence/xs:element', namespaces=ns)
    

    
    for child_element in child_elements:
        if (child_element.tag == '{http://www.w3.org/2001/XMLSchema}element'):
            children.append(child_element)
            
            
    return children


def getattributes(item):
    '''
    Given an lxml lixi element, return its lixi attributes as lxml objects
    '''    
    attributes = []
    if item.tag == '{http://www.w3.org/2001/XMLSchema}complexType':
        child_elements = item.getchildren()[1:]
        for child_element in child_elements:
            if child_element.tag == '{http://www.w3.org/2001/XMLSchema}attribute':
                attributes.append(child_element)   
                
    return attributes


def annotation (schema, documentation, path, label, transactions):
    
    # create new annotation
    new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
    new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')    
    new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
    new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
    new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
    new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')
    
    # build tree
    new_annotation.append (new_documentation)
    new_annotation.append (new_appinfo)
    new_appinfo.append    (new_path)
    new_appinfo.append    (new_label)
    new_appinfo.append    (new_transactions)
    
    # set values
    new_documentation.text = documentation
    new_path.text          = path
    new_label.text         = label
    new_transactions.text  = transactions    
    
    # get the old element
    parent_xpath = '//lx:path[text()="'+path+'"]'
    path_element = schema.xpath(parent_xpath, namespaces=ns)[0]
    element      = path_element.getparent().getparent().getparent()
    
    # swap the annotation
    old_annotation = element.getchildren()[0]
    element.remove(old_annotation)
    element.insert(0, new_annotation)
    
    return element

    
    
def element(schema, name, minOccurs, maxOccurs, type_, documentation, path, label, transactions):
        
    # check if the element exists
    xpath   = '//lx:path[text()="' + path + '"]'
    matches = schema.xpath(xpath, namespaces=ns)
    
    if len(matches) == 0:
        
        print("    element does not exist: " + path) 
        
        # create the new element
        new_element = etree.Element('{http://www.w3.org/2001/XMLSchema}element')
        new_element.set('name', name)
        if type_ != '':
            new_element.set('type', type_)            
        new_element.set('minOccurs', minOccurs)
        new_element.set('maxOccurs', maxOccurs)

        # create child elements 
        new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
        new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
        new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
        new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
        new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
        new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')
        new_complextype   = etree.Element('{http://www.w3.org/2001/XMLSchema}complexType')
        new_sequence      = etree.Element('{http://www.w3.org/2001/XMLSchema}sequence')
         
        # build tree
        new_element.append     (new_annotation)
        new_annotation.append  (new_documentation)
        new_annotation.append  (new_appinfo)
        new_appinfo.append     (new_path)
        new_appinfo.append     (new_label)
        new_appinfo.append     (new_transactions)
        new_element.append     (new_complextype)
        new_complextype.append (new_sequence)
        
        # set values
        new_documentation.text = documentation
        new_path.text          = path
        new_label.text         = label
        new_transactions.text  = transactions
        
        # get parent element
        parent_path = re.sub('.%s$' % name, '', path)
        parent_xpath = '//lx:path[text()="'+parent_path+'"]'
        parent_path_element = schema.xpath(parent_xpath, namespaces=ns)[0]
        parent_element = parent_path_element.getparent().getparent().getparent()
        complextype_element = parent_element.getchildren()[1]
        if (complextype_element.tag == '{http://www.w3.org/2001/XMLSchema}sequence'):
            sequence_element    = complextype_element
        else:
            if (len(complextype_element.getchildren()) == 0):
                sequence_element = etree.Element('{http://www.w3.org/2001/XMLSchema}sequence')
                complextype_element.append(sequence_element)
            else:
                sequence_element    = complextype_element.getchildren()[0]    
            if   (sequence_element.tag == '{http://www.w3.org/2001/XMLSchema}choice'):
                pass
            elif (sequence_element.tag != '{http://www.w3.org/2001/XMLSchema}sequence'):
                complextype_element.insert(0, etree.Element('{http://www.w3.org/2001/XMLSchema}sequence'))
                sequence_element    = complextype_element.getchildren()[0]   
                
        # insert in correct place
        siblings = sequence_element.getchildren()
        if len(siblings) == 0:
            sequence_element.append(new_element)
        else:
            found = False
            for sibling in siblings:
                if sibling.attrib['name'].lower() > new_element.attrib['name'].lower():                
                    sequence_element.insert((sequence_element.index(sibling)), new_element)
                    found = True
                    break
            if not found:
                sequence_element.append(new_element)
                
        removesequences(schema)

        result = new_element
        
    # if we are updating an existing element
    else:
        
        print("    element does exist: " + path) 
        
        old_element = matches[0].getparent().getparent().getparent()
        if path != 'Package':     
            old_element.attrib['minOccurs'] = minOccurs
            old_element.attrib['maxOccurs'] = maxOccurs

        # delete annotation
        old_element.remove(old_element.getchildren()[0])
    
        # create new annotation
        new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
        new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
        new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
        new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
        new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
        new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')
    
        # build tree
        old_element.insert     (0, new_annotation)
        new_annotation.append  (new_documentation)
        new_annotation.append  (new_appinfo)
        new_appinfo.append     (new_path)
        new_appinfo.append     (new_label)
        new_appinfo.append     (new_transactions)
    
        # set values
        new_documentation.text = documentation
        new_path.text          = path
        new_label.text         = label
        new_transactions.text  = transactions        
    
        result = old_element                

    return result
  

def enumeration(schema, name, documentation, path, label, transactions):
    
    # get parent element
    path_with_dot = path.replace("[-]",".") # some enumeration paths have had a "." replace with "[-]" e.g. "insurerList.Smart Business Insurance Pty Ltd[-]"
    parent_path = path_with_dot.replace('.'+name, '')
    parent_xpath = '//lx:path[text()="'+parent_path+'"]'
    parent_path_element = schema.xpath(parent_xpath, namespaces=ns)[0]
    parent_element = parent_path_element.getparent().getparent().getparent()  
    restriction_element = parent_element.getchildren()[1]  
    
    # get enumerations of this list
    siblings = restriction_element.getchildren()

    # check if the enumeration already exists (on the basis of its name)
    simple_type_already_exists = name in names(siblings)    
    
    if simple_type_already_exists:
        # get old element and remove its annotation
        xpath = '//lx:path[text()="'+path+'"]/../../..'
        old_element = schema.xpath(xpath, namespaces=ns)[0]
        old_element.remove(old_element.getchildren()[0])
        new_element = old_element        
    else:
        # create new element
        new_element = etree.Element('{http://www.w3.org/2001/XMLSchema}enumeration')
        new_element.set('value', name)

    # create sub-elements 
    new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
    new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
    new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
    new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
    new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
    new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')

    # build tree
    new_element.append     (new_annotation)
    new_annotation.append  (new_documentation)
    new_annotation.append  (new_appinfo)
    new_appinfo.append     (new_path)
    new_appinfo.append     (new_label)
    new_appinfo.append     (new_transactions)

    # set values
    new_documentation.text = documentation
    new_path.text          = path
    new_label.text         = label
    new_transactions.text  = transactions
    
    # insert in correct place
    
    if len(siblings) == 0:
        restriction_element.append(new_element)
    else:
        found = False
        for sibling in siblings:
            if sibling.attrib['value'].lower() > new_element.attrib['value'].lower():                
                restriction_element.insert((restriction_element.index(sibling)), new_element)
                found = True
                break
        if not found:
            restriction_element.append(new_element)   
            
    return new_element

        
def attribute(schema, name, type_, use, documentation, path, label, transactions):
    
    # check if the element exists
    xpath   = '//lx:path[text()="' + path + '"]'
    matches = schema.xpath(xpath, namespaces=ns)
    
    if len(matches) == 0:
        
        print("    attribute does not exist: " + path)
        
        # get parent element
        parent_path = re.sub('.%s$' % name, '', path)
        parent_xpath = '//lx:path[text()="'+parent_path+'"]'
        parent_path_element = schema.xpath(parent_xpath, namespaces=ns)[0]
        parent_element = parent_path_element.getparent().getparent().getparent()  

        # ###################################
        # Create a new attribute
        # ###################################
               
        new_element = etree.Element('{http://www.w3.org/2001/XMLSchema}attribute')
        new_element.set('name', name)
        new_element.set('type', type_)
        new_element.set('use', use)
        
        # create elements 
        new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
        new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
        new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
        new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
        new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
        new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')
         
        # build tree
        new_element.append     (new_annotation)
        new_annotation.append  (new_documentation)
        new_annotation.append  (new_appinfo)
        new_appinfo.append     (new_path)
        new_appinfo.append     (new_label)
        new_appinfo.append     (new_transactions)
                
        # set values
        new_documentation.text = documentation
        new_path.text          = path
        new_label.text         = label
        new_transactions.text  = transactions
        
        # ##############################
        # get the element to which the attributes will be appended 
        # ##############################
        
        # #########################
        # if it is a complex type (i.e. SignatureType) 
        # the attributes are appended to the "xs:complexType" element 
        # as a sibling to, and after, the annotation and the sequence
        # #########################
        if str(parent_element.tag) == "{http://www.w3.org/2001/XMLSchema}complexType":
            append_to = parent_element   # this is the element to append the attributes to
            siblings  = append_to.getchildren()[2:]    # the annotation and the sequence we only care about the attributes
            indent    = 2   # remember to append after the annotation and the sequence
            
        # #########################
        # if it is a normal element
        # the attributes are appended to the "xs:complexType" element 
        # which is a child of the element element
        # after the sequence element (if there is one)
        # #########################       
        elif str(parent_element.tag) == "{http://www.w3.org/2001/XMLSchema}element":
            children = parent_element.getchildren() # children should be "annotation" and "complexType"
            if children[1].tag == "{http://www.w3.org/2001/XMLSchema}complexType": # check that is is a complex type element where expected
                append_to = children[1]
                siblings  = append_to.getchildren()
                if len(append_to) > 0 and append_to[0].tag == "{http://www.w3.org/2001/XMLSchema}sequence": # if the LIXI element has LIXI child elements
                    siblings = siblings[1:]  # ignore the first one (the sequence) we don't care about LIXI elements only the attributes
                    indent = 1  # remember to insert after the sequence (index 1)
                else:
                    indent = 0  # don't worry about inserting  after the sequence
            else:
                
                raise Exception ("Expected 'complextType' got: " + str(children[1].tag))
        else:
            raise Exception("Expected type of: 'complexType' or 'element' instead got: " + str(parent_element.tag))
 
        
        # #####################################
        # if there are no attributes here, go ahead and add the new one
        # #####################################
        if len(siblings) == 0:
            append_to.insert(indent, new_element)
        
        else: 
            # get the name of the new attribute
            new_element_name = new_element.attrib['name'].lower()
            
            # if the first attribute is higher alphabetically, add the new one to the start
            if siblings[0].attrib['name'].lower() > new_element_name:   
                append_to.insert(indent, new_element)
                
            # get the position of the new attribute
            else:  
                found = False
                for sibling in siblings:
                    if  sibling.attrib['name'].lower() > new_element_name:     
                        position = append_to.index(sibling) 
                        append_to.insert((position), new_element)
                        found = True
                        break
                    
                # otherwise put it on the end
                if not found:
                    append_to.append(new_element)   
                    
        result = new_element
    
    # ###########################################                
    # if we are updating an existing element
    # ###########################################
    else:
        
        print("    attribute does exist")
        
        old_element = matches[0].getparent().getparent().getparent()
        old_element.attrib['type'] = type_
        old_element.attrib['use']  = use
        
        # delete annotation
        old_element.remove(old_element.getchildren()[0])
        
        # create new annotation
        new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
        new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
        new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
        new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
        new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
        new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')
    
        # build tree
        old_element.append     (new_annotation)
        new_annotation.append  (new_documentation)
        new_annotation.append  (new_appinfo)
        new_appinfo.append     (new_path)
        new_appinfo.append     (new_label)
        new_appinfo.append     (new_transactions)
    
        # set values
        new_documentation.text = documentation
        new_path.text          = path
        new_label.text         = label
        new_transactions.text  = transactions        
    
        result = old_element        
                
    return result
            
        
            
def simpletype (schema, name, documentation, path, label, transactions):
    
    # get all the simple types
    simpleTypesAll = schema.xpath("./xs:simpleType", namespaces=ns)
    
    # check if the simple type already exists (on the basis of its name)
    simple_type_already_exists = name in names(simpleTypesAll)
    
    if simple_type_already_exists:
        old_element = schema.xpath("./xs:simpleType[@name='" + name + "']", namespaces=ns)[0]
        # remove the old annotations
        old_element.remove(old_element.getchildren()[0])
        new_element = old_element
        
    else:
        # create new simple type element
        new_element = etree.Element('{http://www.w3.org/2001/XMLSchema}simpleType')
        new_element.set('name', name)        

    # create child elements 
    new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
    new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
    new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
    new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
    new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
    new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')
   
    # build the tree
    new_element.insert (0, new_annotation)
    new_annotation.append  (new_documentation)
    new_annotation.append  (new_appinfo)
    new_appinfo.append     (new_path)
    new_appinfo.append     (new_label)
    new_appinfo.append     (new_transactions)
    
    # set values
    new_documentation.text = documentation
    new_path.text          = path
    new_label.text         = label
    new_transactions.text  = transactions
    
    # if the simple type is new, append a restriction element
    if not simple_type_already_exists:
        new_restriction   = etree.Element('{http://www.w3.org/2001/XMLSchema}restriction')
        new_element.append     (new_restriction)
        new_restriction.set('base','xs:token')
    
    # sort the simple type into the three categories: 'list' 'pattern' and 'type' 
    simpleTypesLists = []
    simpleTypesPatterns = []
    simpleTypesTypes = []     
    
    # add the other elements to the correct lists
    for simpleType in simpleTypesAll:
        name_ = simpleType.attrib['name'].lower()
        if name_.endswith('list'):
            simpleTypesLists.append(simpleType)  
        elif name_.endswith('pattern'):
            simpleTypesPatterns.append(simpleType) 
        elif name_.endswith('type'):
            simpleTypesTypes.append(simpleType)
            
    # get position of new item
    position = len(schema.xpath("./xs:complexType", namespaces=ns)) + 2
    
    if name.endswith('List'):
        for sibling in simpleTypesLists:
            if sibling.attrib['name'].lower() > new_element.attrib['name'].lower():                
                break
            position += 1
            
    if name.endswith('Pattern'):
        position += len(simpleTypesLists)
        for sibling in simpleTypesPatterns:
            if sibling.attrib['name'].lower() > new_element.attrib['name'].lower():                
                break
            position += 1        
            
    if name.endswith('Type'):
        position += len(simpleTypesLists) + len(simpleTypesPatterns)
        for sibling in simpleTypesTypes:
            if sibling.attrib['name'].lower() > new_element.attrib['name'].lower():                
                break
            position += 1          
        
    schema.insert(position, new_element)    
    
    return new_element


def put_complextype (schema, name, documentation, path, label, transactions):
    


    # get all the complex types
    complexTypesAll = schema.xpath("./xs:complexType", namespaces=ns)

    # if the complex type already exists
    if name in names(complexTypesAll):
        
        # get the old complex type element
        old_element = schema.xpath("./xs:complexType[@name='" + name + "']", namespaces=ns)[0]
        
        # delete the old annotation
        old_element.remove(old_element.getchildren()[0]) 
        
        # create child elements 
        new_annotation    = etree.Element('{http://www.w3.org/2001/XMLSchema}annotation')
        new_documentation = etree.Element('{http://www.w3.org/2001/XMLSchema}documentation')
        new_appinfo       = etree.Element('{http://www.w3.org/2001/XMLSchema}appinfo')
        new_path          = etree.Element('{lixi.org.au/schema/appinfo_elements}path')
        new_label         = etree.Element('{lixi.org.au/schema/appinfo_elements}label')
        new_transactions  = etree.Element('{lixi.org.au/schema/appinfo_instructions}transactions')        
        
        # build tree
        old_element.insert     (0, new_annotation)
        new_annotation.append  (new_documentation)
        new_annotation.append  (new_appinfo)
        new_appinfo.append     (new_path)
        new_appinfo.append     (new_label)
        new_appinfo.append     (new_transactions)
        
        # set values
        new_documentation.text = documentation
        new_path.text          = path
        new_label.text         = label
        new_transactions.text  = transactions        
        
        new_element = old_element
        
    else:
        
        # create new complex type element
        new_element = etree.Element('{http://www.w3.org/2001/XMLSchema}complexType')
        new_element.set('name', name)
        
        raise Exception ("The function to add a new complex type has not been finalised.")
        
    return new_element
 
    
        
  
def removesequences(schema):      
    # remove empty sequences
    empty_sequences = schema.xpath('//xs:sequence[not(*)]', namespaces=ns)
    for empty in empty_sequences:
        empty.getparent().remove(empty)


def removecomplextypes(schema):         
    # remove empty complextype
    empty_complexType = schema.xpath('//xs:complexType[not(*)]', namespaces=ns)
    for empty in empty_complexType:
        empty.getparent().remove(empty)    


def main(annotated_unexploded_schema):

    
    schema       = annotated_unexploded_schema
    elements     = get_all_lixi_elements(schema)
    complexTypes = get_all_lixi_complex_types(schema)
    
    
    
    rootelement  = 'Package'
    annotations  = True
    update_paths = True
    
    for complextype in complexTypes:
        complextype_name = get_name(complextype)
        elements_using_the_type = get_all_lixi_elements_using_a_complex_type(schema, complextype)
        
        for element_using_the_type in elements_using_the_type:
            
            
            
            minoccurs_of_element_using_the_type = minOccurs([element_using_the_type])[0]
            maxoccurs_of_element_using_the_type = maxOccurs([element_using_the_type])[0]
            
            if maxoccurs_of_element_using_the_type == "": 
                maxoccurs_of_element_using_the_type = "1"
    
            cloned_complextype             = clone(complextype)
            cloned_complextype_paths       = get_all_lixi_paths(cloned_complextype)
            
            
            if update_paths:
                name_of_element_using_the_type = get_name(element_using_the_type)
                path_of_element_using_the_type = get_path(element_using_the_type)        
        
                for path in cloned_complextype_paths:
                    path.text = path.text.replace(complextype_name, path_of_element_using_the_type) 
                
            if annotations:
                cloned_complextype_children = cloned_complextype.getchildren()[1:]
                
            else: 
                cloned_complextype_children = cloned_complextype.getchildren()
            
            # add a '<xs:complexType' element under the element using the complex type
            
            new_complextype_element = etree.Element('{http://www.w3.org/2001/XMLSchema}complexType')
            
    
            element_using_the_type.append(new_complextype_element)
            
            for child in cloned_complextype_children:
                new_complextype_element.append(child)
                
            # remove the 'type' attribute from the element using the complex type
            element_using_the_type.attrib.pop('type')
            
            element_using_the_type.attrib['minOccurs'] = minoccurs_of_element_using_the_type
            
            element_using_the_type.attrib['maxOccurs'] = maxoccurs_of_element_using_the_type
            
        complextype.getparent().remove(complextype) 
        
    root_elements = schema.getchildren()
        
    for child in root_elements:
        
        if 'name' in child.attrib and not child.attrib['name'] == rootelement and not child.tag == "{http://www.w3.org/2001/XMLSchema}simpleType" :
    
            child.getparent().remove(child)
                
    annotated_exploded_schema = schema
                
    return annotated_exploded_schema  


import os, glob, re, isodate, datetime, pkg_resources, io, json
from decimal import *
from lxml import etree as elemTree
from xmljson import BadgerFish
from collections import OrderedDict
from operator import attrgetter

if __name__ == "_LIXI":
    import _explode_schema
else:
    from lixi import (
        _explode_schema
    )

ns = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "lx": "lixi.org.au/schema/appinfo_elements",
    "li": "lixi.org.au/schema/appinfo_instructions",
}


def to_json(xml_etree, xsd_annotated_schema):
    """ Converts a Python XML etree to JSON etree and arranges the result per schema.
    
    Args:
        xml_etree (:obj:`lxml.etree`, required): XML Message to convert to JSON message.
        xsd_annotated_schema (:obj:`lxml.etree`, required): A LIXI XML Schema.
    
    Returns:
        A python JSON dict object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    json_object = {}

    def json_verify_types(json_data, xsd_element):
        elementsandatt_mappings = {}
        complex_mappings = []
        simpletype_mappings = {}
        xml_to_json_message = """{
                "anyType":  "string",
                    "anySimpleType":  "string",
                    "string":  "string",
                    "normalizedString":  "string",
                    "token":  "string",
                    "language":  "string",
                    "Name":  "string",
                    "NCName":  "string",
                    "ID":  "id",
                    "IDREF":  "idref",
                    "IDREFS":  "idref",
                    "ENTITY":  "string",
                    "ENTITIES":  "string",
                    "NMTOKEN":  "string",
                    "NMTOKENS": "string",
                    "boolean": "boolean",
                    "base64Binary": "base64",
                    "hexBinary": "string",
                    "float": "number",
                    "decimal": "number",
                    "integer": "integer",
                    "nonPositiveInteger": "integer",
                    "negativeInteger": "integer",
                    "long": "integer",
                    "int": "integer",
                    "short": "integer",
                    "byte": "integer",
                    "nonNegativeInteger": "integer",
                    "unsignedLong": "integer",
                    "unsignedInt": "integer",
                    "unsignedShort": "integer",
                    "unsignedByte": "integer",
                    "positiveInteger": "integer",
                    "double": "number",
                    "anyURI": "string",
                    "QName": "string",
                    "NOTATION": "string",
                    "duration": "string",
                    "dateTime": "datetime",
                    "date": "date",
                    "time": "string",
                    "gYearMonth": "string",
                    "gYear": "year",
                    "gMonthDay": "string",
                    "gDay": "string",
                    "gMonth": "string"
    }"""

        # reading and storing the native XMLtypes to their equivalent json types
        type_mappings = json.loads(xml_to_json_message)

        # reading and storing the simpleTypes to their specified base XMLtypes
        for simple_type in xsd_element.findall("./xs:simpleType", namespaces=ns):
            element_name = simple_type.attrib.get("name")
            element_type = simple_type.find("xs:restriction", ns).attrib["base"]
            if "xs:" in element_type:
                element_type = element_type[3:]

            simpletype_mappings[element_name] = element_type

        # reading and storing the attributes and info containing elements to their simpletypes
        for every_element in xsd_element.iter():
            if "www.w3.org" in every_element.tag:
                try:
                    if (
                        "attribute" in every_element.tag
                        and every_element.attrib["type"] is not None
                    ):
                        element_name = "@" + every_element.attrib.get("name")
                        element_type = every_element.attrib["type"]
                        elementsandatt_mappings[element_name] = element_type
                    elif (
                        "element" in every_element.tag
                        and every_element.attrib["type"] is not None
                        and every_element.attrib.get("name") is not None
                    ):
                        element_name = "$" + every_element.attrib.get("name")
                        element_type = every_element.attrib["type"]
                        elementsandatt_mappings[element_name] = element_type
                except Exception as e:
                    pass

        # reading and removing the referance to the complexType as they will be meaningless in a json message file
        for complex_type in xsd_element.findall("./xs:complexType", namespaces=ns):
            element_name = complex_type.attrib.get("name")
            complex_mappings.append(element_name)

        for del_elementname in complex_mappings:
            elementsandatt_mappings = {
                k: v for k, v in elementsandatt_mappings.items() if v != del_elementname
            }

        # now we read and replace the values in elementsandatt_mappings with values of simpletype_mappings
        for k, v in elementsandatt_mappings.items():
            elementsandatt_mappings[k] = simpletype_mappings[v]

        # now we read and replace the values in elementsandatt_mappings with values of type_mappings
        for k, v in elementsandatt_mappings.items():
            elementsandatt_mappings[k] = type_mappings[v]

        # Voila! - now we have a list which has a type attached with a element and attribute name
        json_change_type(json_data, elementsandatt_mappings)

    def json_change_type(json_dict, elementsandatt_mappings):
        for k, v in json_dict.items():
            if isinstance(v, dict):
                json_change_type(v, elementsandatt_mappings)
            elif isinstance(v, list):
                for json_obj in v:
                    json_change_type(json_obj, elementsandatt_mappings)
            else:
                if k != "$":
                    type_cast_value = None
                    lookup = None
                    storage = None

                    if k in elementsandatt_mappings:
                        type_cast_value = v
                        lookup = elementsandatt_mappings[k]
                        storage = json_dict[k]
                    elif "$" + k in elementsandatt_mappings:
                        if "$" not in json_dict[k]:
                            json_dict[k]["$"] = ""

                        type_cast_value = json_dict[k]["$"]
                        lookup = elementsandatt_mappings["$" + k]
                        storage = json_dict[k]["$"]

                    if "number" == lookup or "decimal" == lookup or "float" == lookup:
                        type_cast_value = float(type_cast_value)
                    elif "integer" == lookup:
                        type_cast_value = int(type_cast_value)
                    elif "boolean" == lookup:
                        type_cast_value = bool(type_cast_value)
                    elif "base64" == lookup:
                        if not re.match(
                            "^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$",
                            type_cast_value,
                        ):
                            raise Exception("Inline Attachment is not a Base 64 string")
                        type_cast_value = type_cast_value
                    elif "datetime" == lookup:
                        try:
                            dt = isodate.parse_datetime(type_cast_value)
                            type_cast_value = type_cast_value
                        except ValueError as e:
                            raise Exception(
                                "'"
                                + type_cast_value
                                + "' is not a valid ISO date string"
                            )
                    elif "date" == lookup:
                        try:
                            dt = datetime.datetime.strptime(type_cast_value, "%Y-%m-%d")
                            type_cast_value = type_cast_value
                        except ValueError as e:
                            raise Exception(
                                "'" + type_cast_value + "' is not a valid date"
                            )
                    elif "year" == lookup:
                        try:
                            dt = datetime.datetime.strptime(type_cast_value, "%Y")
                            type_cast_value = type_cast_value
                        except ValueError as e:
                            raise Exception(
                                "'" + type_cast_value + "' is not a valid date"
                            )
                    elif "id" == lookup or "idref" == lookup:
                        if not re.match("^[a-zA-Z0-9-_.]+$", type_cast_value):
                            raise Exception(
                                "ID '"
                                + type_cast_value
                                + "' contains forbidden characters"
                            )
                        elif not re.match("^[a-zA-Z_]+$", type_cast_value[:1]):
                            raise Exception(
                                "ID '"
                                + type_cast_value
                                + "' starts with a forbidden character"
                            )
                        type_cast_value = type_cast_value

                        # no point converting string types back to string // use later for dates and complex types
                    if k in elementsandatt_mappings:
                        json_dict[k] = type_cast_value
                    elif "$" + k in elementsandatt_mappings:
                        json_dict[k]["$"] = type_cast_value

    def json_change_occurance(json_dict, parent, xsd_element):
        for k, v in json_dict.items():
            
            if isinstance(v, dict):            
                json_change_occurance(v, k, xsd_element)
                #if k in list_elements_names:
                #    json_dict[k] = []
                #    json_dict[k].append(v)
                
                if parent!=None:
                    elements = xsd_element.xpath(
                        '//xs:element[@name="'+k+'"]', namespaces=ns
                    )
                    
                    element_found = None
                    for xsd_elem in elements:
                        if xsd_elem.getparent() != None:
                            a = xsd_elem.getparent().getparent().getparent()
                            if 'name' in a.attrib:
                                if a.attrib['name'] == parent:
                                    element_found = xsd_elem
                        
                    if element_found!= None:
                        if 'maxOccurs' in element_found.attrib:
                            if element_found.attrib['maxOccurs'] == 'unbounded':
                                json_dict[k] = []
                                json_dict[k].append(v)                                 
                            elif int(element_found.attrib['maxOccurs']) > 1:
                                json_dict[k] = []
                                json_dict[k].append(v)                                

            elif isinstance(v, list):
                for json_obj in v:
                    json_change_occurance(json_obj, k, xsd_element)

    

    bf = BadgerFish(xml_fromstring=False)
    json_object = bf.data(xml_etree)
    
    # Converts json types from string types to appropriate types specified in XML Schema
    json_verify_types(json_object, xsd_annotated_schema)
    # Converts elements that are single json objects in message (but repeatable in XML schema) to list/arrays
    exploded_schema = _explode_schema.main(xsd_annotated_schema)
    json_change_occurance(json_object, None, exploded_schema)
      
    return json_object


def to_json_string(xml_str, xsd_annotated_schema_str):
    """ Converts a Python XML string to JSON string and arranges the result per schema.
    
    Args:
        xml_str (:obj:`str`, required): XML Message string to convert to JSON message.
        xsd_annotated_schema_str (:obj:`str`, required): A LIXI XML Schema string.
    
    Returns:
        A python JSON string message object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    xml_etree = elemTree.fromstring(xml_str)
    xsd_annotated_schema = elemTree.fromstring(xsd_annotated_schema_str)

    converted_json = to_json(xml_etree, xsd_annotated_schema)
    return json.dumps(converted_json, sort_keys=True, indent=4, ensure_ascii=False)


def to_xml(json_dict, xsd_annotated_schema):
    """ Converts a Python JSON dict to XML etree and arranges the result per schema.
    
    Args:
        json_dict (:obj:`dict`, required): JSON Message string to convert to XML message.
        xsd_annotated_schema (:obj:`lxml.etree`, required): A LIXI XML Schema string.
    
    Returns:
        A python XML Etree object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    def xml_get_sorted_list(schema_root):
        sorting_elem_list = {}
        sorting_list = {}
        sorting_complex_list = {}

        for elem in schema_root.iter(tag=("{*}element", "{*}complexType")):

            if "element" in elem.tag and elem.attrib.get("name") is not None:
                if elem.attrib.get("type") is None:
                    key_name = (
                        str(
                            elem.xpath(
                                "./xs:annotation/xs:appinfo/lx:path/text()",
                                namespaces=ns,
                            )
                        )
                        .strip("[]'")
                        .replace(".", "/")
                    )

                    if (
                        elem.find(
                            "./xs:complexType/xs:sequence/xs:element", namespaces=ns
                        )
                        is not None
                    ):
                        sorting_elem_list[key_name] = []
                        for child_xsd_element in elem.xpath(
                            "./xs:complexType/xs:sequence/xs:element", namespaces=ns
                        ):
                            sorting_elem_list[key_name].append(
                                child_xsd_element.attrib.get("name")
                            )

                    elif (
                        elem.find(
                            "./xs:complexType/xs:choice/xs:element", namespaces=ns
                        )
                        is not None
                    ):
                        sorting_elem_list[key_name] = []
                        for child_xsd_element in elem.findall(
                            "./xs:complexType/xs:choice/xs:element", namespaces=ns
                        ):
                            sorting_elem_list[key_name].append(
                                child_xsd_element.attrib.get("name")
                            )

                elif elem.attrib.get("type") is not None:
                    key_name = (
                        str(
                            elem.xpath(
                                "./xs:annotation/xs:appinfo/lx:path/text()",
                                namespaces=ns,
                            )
                        )
                        .strip("[]'")
                        .replace(".", "/")
                    )
                    sorting_elem_list[key_name + "/" + elem.attrib.get("type")] = []

            elif "complexType" in elem.tag:
                if elem.attrib.get("name") is not None:
                    key_name = (
                        str(
                            elem.xpath(
                                "./xs:annotation/xs:appinfo/lx:path/text()",
                                namespaces=ns,
                            )
                        )
                        .strip("[]'")
                        .replace(".", "/")
                    )

                    if elem.find("./xs:sequence/xs:element", namespaces=ns) is not None:
                        sorting_complex_list[key_name] = []
                        for child_xsd_element in elem.xpath(
                            "./xs:sequence/xs:element", namespaces=ns
                        ):
                            sorting_complex_list[key_name].append(
                                child_xsd_element.attrib.get("name")
                            )

        for k, v in sorting_complex_list.items():
            for key, value in sorting_elem_list.items():
                if k in key:
                    sorting_list["/" + key.replace("/" + k, "")] = v

        for k, v in sorting_elem_list.items():
            if len(v) > 0:
                sorting_list["/" + k] = v

        return sorting_list

    def xml_sort_message(unsorted_xml, sorting_list):

        tree = elemTree.ElementTree(unsorted_xml)

        for e in unsorted_xml.xpath("//*"):

            keys = e.attrib
            sorted_attrib = OrderedDict(sorted(keys.items()))
            e.attrib.clear()
            e.attrib.update(sorted_attrib)

            elem_path_key = tree.getpath(e).strip("[0123456789]")

            if elem_path_key in sorting_list:
                e[:] = sorted(
                    e, key=lambda elem: get_elem_key(elem, elem_path_key, sorting_list)
                )

        return unsorted_xml

    def get_elem_key(elem, elem_path_key, sorting_list):

        try:
            if sorting_list[elem_path_key].index(elem.tag) is not None:
                return sorting_list[elem_path_key].index(elem.tag)
            else:
                return 0
        except:
            return 0

    bf = BadgerFish()
    unsorted_xml = bf.etree(json_dict)[0]

    sorting_list = xml_get_sorted_list(xsd_annotated_schema)
    converted_sorted_xml = xml_sort_message(unsorted_xml, sorting_list)

    return converted_sorted_xml


def to_xml_string(json_str, xsd_annotated_schema_str):
    """ Converts a Python JSON string to XML string and arranges the result per schema.
    
    Args:
        json_str (:obj:`str`, required): JSON Message string to convert to XML message.
        xsd_annotated_schema_str (:obj:`str`, required): A LIXI XML Schema string.
    
    Returns:
        A python XML string message object.
    
    Raises:
        LIXIResouceNotFoundError: If the schema is not found at the schema path.
        LIXIInvalidSyntax: If the schema file is not well formed.
    """

    json_dict = json.loads(json_str)
    xsd_annotated_schema = elemTree.fromstring(xsd_annotated_schema_str)

    converted_xml = to_xml(json_dict, xsd_annotated_schema)
    return elemTree.tostring(converted_xml, pretty_print=True).decode("utf-8")
